#include <time.h>
#include <ctype.h>	// tolower()
#include <stdio.h>
#include <string.h>
#ifndef _WIN32
#include <unistd.h>
#include <stdlib.h>
#else
#include <windows.h>
#endif
#include <FL/Fl.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Pixmap.H>
#include <FL/Fl_Box.H>
#include <FL/fl_draw.H>

/************************************************************************

    nixieclock - a nixie desktop timepiece
    Copyright (C) 2000 Greg Ercolano, erco@3dsite.com

    REVISIONS
    ---------
    erco@netcom.com 1.10 11/25/00 -- added composite digits
    erco@3dsite.com 1.20 07/19/01 -- OPENGL version, renamed nixiedate
    erco@3dsite.com 1.30 07/28/01 -- Added entire nixie alphabet
    erco@3dsite.com 1.40 02/02/02 -- Rewrote using fltk image drawing, cleanup
    erco@3dsite.com 1.41 02/05/02 -- Efficient redraws/holdtime now .5 secs

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

// THE DIGIT XPM FILES
#include "xpms/invalid.xpm"
#include "xpms/space.xpm"
#include "xpms/slash.xpm"
#include "xpms/dots.xpm"
#include "xpms/dash.xpm"
#include "xpms/d0.xpm"
#include "xpms/d1.xpm"
#include "xpms/d2.xpm"
#include "xpms/d3.xpm"
#include "xpms/d4.xpm"
#include "xpms/d5.xpm"
#include "xpms/d6.xpm"
#include "xpms/d7.xpm"
#include "xpms/d8.xpm"
#include "xpms/d9.xpm"
#include "xpms/a_a.xpm"
#include "xpms/a_b.xpm"
#include "xpms/a_c.xpm"
#include "xpms/a_d.xpm"
#include "xpms/a_e.xpm"
#include "xpms/a_f.xpm"
#include "xpms/a_g.xpm"
#include "xpms/a_h.xpm"
#include "xpms/a_i.xpm"
#include "xpms/a_j.xpm"
#include "xpms/a_k.xpm"
#include "xpms/a_l.xpm"
#include "xpms/a_m.xpm"
#include "xpms/a_n.xpm"
#include "xpms/a_o.xpm"
#include "xpms/a_p.xpm"
#include "xpms/a_q.xpm"
#include "xpms/a_r.xpm"
#include "xpms/a_s.xpm"
#include "xpms/a_t.xpm"
#include "xpms/a_u.xpm"
#include "xpms/a_v.xpm"
#include "xpms/a_w.xpm"
#include "xpms/a_x.xpm"
#include "xpms/a_y.xpm"
#include "xpms/a_z.xpm"

#define UPDATESECS	20.0
#define COMPHOLDSECS  	0.5

#define XMARGIN		5		// margin around clock on X
#define YMARGIN		5		// margin around clock on Y
#define FONTWIDTH	23
#define FONTHEIGHT	31

// CLASS OVERLOADS Fl_Double_Window
class MyWindow : public Fl_Double_Window
{
    int handle(int e);			// event handler method
    void draw(void);			// screen draw method
public:
    MyWindow(int x,int y,int w,int h,const char *l=0) : Fl_Double_Window(x,y,w,h,l) 
        { box(FL_FLAT_BOX); color(FL_BLACK); }
    MyWindow(int w,int h,const char *l=0):Fl_Double_Window(w,h,l) 
        { box(FL_FLAT_BOX); color(FL_BLACK); }
};

// IMAGE OF A DIGIT
//     Imports an XPM file.
//     Keeps track of the resulting RGB array suitable for glWritePixels(),
//     and maintains the width and height.
//
class Digit
{
public:
    char c;				// character's ascii value
    int w, h;				// width/height of digit
    unsigned char *pixels;		// RGB array of pixels (w*h*4)

    Digit();
    Digit(char ascii, char **xpm);	// ctor imports xpm
    ~Digit();

    void Comp(Digit&, Digit&);		// comp two different digits into class
};

// GLOBALS
MyWindow  *G_mainwin;				// main window
static char 
           G_dig[40],				// digit values for each position
           G_lastdig[40],			// last digits (to detect deltas)
	   G_firsttime = 1,
           G_test = 0,				// 1=run alphanumeric test
	  *G_timestring = "%H:%M %a %b %d";	// default "23:59 Sat Dec 31"

Digit     *G_digits[64];			// rgb digit array
int        G_tdigits = 0;			// total digits loaded
int        G_dirty = 0;				// flag if last redraw needs to be wiped
time_t	   G_countdown = 0;			// countdown time if non-zero


// CONVERT XPM TO RGB
unsigned char *Xpm2Rgb(char **xpm, int& w, int& h)
{
    static long vals[256], val;
    int t, x, y, colors, cpp;
    unsigned char c;
    unsigned char *savergb, *rgb;

    // PARSE HEADER
    if ( sscanf(xpm[0], "%d%d%d%d", &w, &h, &colors, &cpp) != 4 )
        { fprintf(stderr, "nixieclock: bad xpm header: '%s'\n", xpm[0]); exit(1); }

    savergb = rgb = (unsigned char*)malloc(h * w * 3);

    // LOAD XPM COLORMAP LONG ENOUGH TO DO CONVERSION
    for ( t=0; t<colors; t++ )
    {
	sscanf(xpm[t+1], "%c c #%lx", &c, &val);
	vals[c] = val;
    }

    // COLORMAP -> RGB CONVERSION
    //    Get low 3 bytes from vals[]
    //
    char *p;
    for ( y=h-1; y>0; y-- )
    {
	for ( p = xpm[1+colors+(h-y-1)], x=0; x<w; x++, rgb+=3 )
	{
	    val = vals[*p++];
	    *(rgb+2) = val & 0xff; val >>= 8;  // 2:B
	    *(rgb+1) = val & 0xff; val >>= 8;  // 1:G
	    *(rgb+0) = val & 0xff;             // 0:R
	}
    }
    return(savergb);
}

// CTOR
Digit::Digit()
{
    pixels = NULL;
    w = 0;
    h = 0;
    c = 0;
}

// CONVERT ASCII/XPM -> RGB IMAGE SUITABLE FOR WRITEPIXELS
Digit::Digit(char ascii, char **xpm)
{
    c = ascii;
    pixels = Xpm2Rgb(xpm, w, h);
}

// FREE THE DIGIT IMAGE
Digit::~Digit()
{
    if ( pixels ) free(pixels);
}

// COMP ONE DIGIT OVER ANOTHER
//    Both input digits must be the same size.
//    Neither digit's pixels are affected; a copy is made.
//
void Digit::Comp(Digit& a, Digit& b)
{
    if ( pixels ) free(pixels);

    int size = a.w * a.h * 3;
    unsigned char 
         *ap = a.pixels,
         *bp = b.pixels,
	 *rgb;

    // GET WIDTH, HEIGHT, PIXELS
    w = a.w;
    h = a.h;
    pixels = rgb = (unsigned char*)malloc(size);

    // QUICK COMP
    for ( int t=0; t<size; t++ )
        *rgb++ = ( *ap++ | *bp++ );		// logical 'OR' sufficient
}

// UPDATE TIME
//     If nothing changed, don't redraw
//
int UpdateTime()
{
    // GET CURRENT DATE/TIME AS SEPARATE DIGITS
    time_t tim = time(NULL);
    struct tm *buf;
    if (G_countdown) 
    {
        // epoch countdown
        tim = G_countdown - tim;
	tim = ( tim < 0 ) ? 0 : tim; 	// stop counter at zero
	buf = gmtime(&tim);
    }
    else
    {
	buf = localtime(&tim);
    }
    strftime(G_dig, 40, G_timestring, buf);

    // FIRST TIME? SHORT
    if ( G_firsttime )
        { G_firsttime = 0; return(1); }

    // ANY CHANGES?
    return(strcmp(G_dig, G_lastdig) != 0);
}

// HANDLE EVENTS IN THE WINDOW
int MyWindow::handle(int e)
{
    static int xoff = 0, yoff = 0;
    switch ( e )
    {
        // DOWNCLICK IN WINDOW CREATES CURSOR OFFSETS
	case FL_PUSH:
	    xoff = x() - Fl::event_x_root();
	    yoff = y() - Fl::event_y_root();
	    return(1);

	case FL_RELEASE:
	    show();		// raise
	    return(1);

	case FL_DRAG:
	    // DRAG THE WINDOW AROUND THE SCREEN
	    position(xoff + Fl::event_x_root(), yoff + Fl::event_y_root());
	    UpdateTime();
	    redraw();
	    return(1);
    }
    return(Fl_Double_Window::handle(e)); 
}

// NUMERIC VALUE (0-9) RETURNS APPROPRIATE RGB DIGIT
Digit *Value2Digit(int val)
{
    for ( int t=0; t<G_tdigits; t++ )
        if ( G_digits[t]->c == tolower(val) )		// force lowercase
	    return(G_digits[t]);
    return(G_digits[0]);				// unknown character
}

// DRAW TWO DIGITS COMPED OVER ONE ANOTHER AT X/Y POSITION
//    Returns 1 if comp occurred.
//
int DrawDigitComp(int aval, int bval, int x, int y)
{
    // DONT COMP UNLESS DIGITS ARE DIFFERENT
    if ( aval == bval )
    {
	Digit *a = Value2Digit(aval);
	fl_draw_image(a->pixels, x, y, a->w, a->h-1);
	return(0);
    }
    else
    {
	static Digit tmp;
	Digit *a = Value2Digit(aval),
	      *b = Value2Digit(bval);
	tmp.Comp(*a, *b);
	fl_draw_image(tmp.pixels, x, y, tmp.w, tmp.h-1);
	return(1);
    }
}

// TIMER: CAUSES A REDRAW THAT UPDATES THE TIME
void Tick_CB(void *) 
{
    float secs = ( G_test ) ? 1.0 : COMPHOLDSECS;

    // TIME CHANGED? REDRAW
    int change = UpdateTime();

    Fl::repeat_timeout(secs, Tick_CB);

    // REDRAW IF TIME CHANGED 
    //    ..or if last draw we did a comp, in which case 
    //    we need to draw again to clear the 'comp image'.
    //
    if ( G_dirty || change )
	G_mainwin->redraw();
}

// WINDOW DRAW HANDLER
void MyWindow::draw(void)
{
    static int testindex = 0;

    Fl_Double_Window::draw();

    // DRAW DIGITS
    //    Overlap ones that changed
    //
    int x = XMARGIN;
    G_dirty = 0;
    for ( int t=0; G_dig[t]; t++ )
    {
        if ( G_test )
	    G_dirty |= 
	        DrawDigitComp(G_digits[testindex]->c, G_digits[testindex]->c, 
			      x, YMARGIN);
        else
	    G_dirty |=
		DrawDigitComp(G_dig[t], G_lastdig[t], x, YMARGIN);

	x += FONTWIDTH + 2;
	G_lastdig[t] = G_dig[t];
    }

    // TEST MODE? CYCLE THROUGH ALL DIGITS
    if ( G_test )
	{ testindex = ( testindex + 1 ) % G_tdigits; }
}

void HelpAndExit()
{
    fprintf(stderr, 
	"USAGE\n"
	"    nixieclock [options]\n"
	"\n"
	"OPTIONS\n"
	"    -d              # digit date mode (eg. 07:57 07-30-01)\n"
	"    -s <timestr>    # use strftime(3) date format\n"
	"    -t              # test pattern for alphanumeric display\n"
	"    -n              # no background mode (don't fork)\n"
	"    -geometry <pos> # sets geometry (x windows style positions)\n"
	"    -c <epochtime>  # counts down to epochtime\n"
	"    -help           # help\n"
	"\n"
	"EXAMPLES\n"
	"    nixieclock                     -- 07:57 Sat Feb 2000\n"
	"    nixieclock -d                  -- 07:57 07-30-01\n"
	"    nixieclock -s '%%H:%%M:%%S'       -- 23:59:59 (counts seconds)\n"
	"    nixieclock -s '%%H:%%M:%%S %%a %%b %%d'  -- 23:59:59 Sat Dec 31\n"
	"    nixieclock -geometry -0-0      -- position lower right\n"
	"    nixieclock -geometry -5-5      -- position 5 pixels lower right\n"
	"    nixieclock -geometry +0-0      -- position lower left\n"
	"\n");
    exit(1);
}

// MAIN
int main(int argc, char **argv) 
{
    // BANNER
    fprintf(stderr, "\nnixieclock %s\n"
                    "(C) Copyright 2002 Greg Ercolano (erco@3dsite.com).\n"
		    "This code is GPL'ed.\n\n", VERSION);

    // PARSE ARGUMENTS
    {
	int nofork = 0, arg_offset = 0;
	for ( int t=1; t<argc; t++ )
	{
	    // HELP
	    if ( strcmp(argv[t], "-help") == 0 ||
	         strcmp(argv[t], "--help") == 0) HelpAndExit();

	    if ( argv[t][0] == '-' )
	    {
		if ( argv[t][0] == 0 ) 
		    HelpAndExit();

		switch ( argv[t][1] )
		{
		    case 't': 
			G_test = 1;
			break;

		    case 'n': 
			nofork = 1; 
			break;

		    case 's':
			G_timestring = argv[++t];
			++arg_offset;
			break;

		    // 23:59
		    case 'h': G_timestring="%H:%M"; break;

		    // 23:59 12/31/99
		    case 'd': G_timestring="%H:%M %D"; break;

		    // 23:59 Sat Dec 31
		    case 'm': G_timestring="%H:%M %a %b %d"; break; 

		    case 'c':
			G_countdown = strtol(argv[++t], NULL, 0);
			++arg_offset;
			break;

		    default: continue;
		}
		++arg_offset;
	    }
	}

	// PREVENT FLTK FROM SEEING ARGUMENTS WE PARSE
	argc -= arg_offset;
	argv += arg_offset;

#ifndef _WIN32
#ifndef DARWIN
	// UNIX: RUN IN THE BACKGROUND
	if ( ! nofork )
	{
	    switch ( fork() )
	    {
		case -1: perror("fork"); exit(1); 	// ERROR
		case  0: break;				// CHILD
		default: exit(0);			// PARENT DISCONNECTS
	    }
	}
#endif
#endif
    }

    // SET VISUAL MODE
    //
    //     FL_RGB prevents color map eating under X windows.
    //
    //     FL_DOUBLE|FL_INDEX enables hardware double buffering,
    //     but looks bad on Indy; digit backgrounds are gray.
    //     Fltk does a good job emulating double buffering
    //     in Fl_Double_Window w/out flickering, so leave off.
    //
    Fl::visual(FL_RGB);

    // PASS OTHER ARGS TO FLTK (-geometry, etc)
    Fl::args(argc, argv);

    // INITIALIZE
    G_tdigits = 0;
    G_digits[G_tdigits++] = new Digit('\x7f', invalid_xpm);
    G_digits[G_tdigits++] = new Digit('a', a_a_xpm);
    G_digits[G_tdigits++] = new Digit('b', a_b_xpm);
    G_digits[G_tdigits++] = new Digit('c', a_c_xpm);
    G_digits[G_tdigits++] = new Digit('d', a_d_xpm);
    G_digits[G_tdigits++] = new Digit('e', a_e_xpm);
    G_digits[G_tdigits++] = new Digit('f', a_f_xpm);
    G_digits[G_tdigits++] = new Digit('g', a_g_xpm);
    G_digits[G_tdigits++] = new Digit('h', a_h_xpm);
    G_digits[G_tdigits++] = new Digit('i', a_i_xpm);
    G_digits[G_tdigits++] = new Digit('j', a_j_xpm);
    G_digits[G_tdigits++] = new Digit('k', a_k_xpm);
    G_digits[G_tdigits++] = new Digit('l', a_l_xpm);
    G_digits[G_tdigits++] = new Digit('m', a_m_xpm);
    G_digits[G_tdigits++] = new Digit('n', a_n_xpm);
    G_digits[G_tdigits++] = new Digit('o', a_o_xpm);
    G_digits[G_tdigits++] = new Digit('p', a_p_xpm);
    G_digits[G_tdigits++] = new Digit('q', a_q_xpm);
    G_digits[G_tdigits++] = new Digit('r', a_r_xpm);
    G_digits[G_tdigits++] = new Digit('s', a_s_xpm);
    G_digits[G_tdigits++] = new Digit('t', a_t_xpm);
    G_digits[G_tdigits++] = new Digit('u', a_u_xpm);
    G_digits[G_tdigits++] = new Digit('v', a_v_xpm);
    G_digits[G_tdigits++] = new Digit('w', a_w_xpm);
    G_digits[G_tdigits++] = new Digit('x', a_x_xpm);
    G_digits[G_tdigits++] = new Digit('y', a_y_xpm);
    G_digits[G_tdigits++] = new Digit('z', a_z_xpm);
    G_digits[G_tdigits++] = new Digit('0', d0);
    G_digits[G_tdigits++] = new Digit('1', d1);
    G_digits[G_tdigits++] = new Digit('2', d2);
    G_digits[G_tdigits++] = new Digit('3', d3);
    G_digits[G_tdigits++] = new Digit('4', d4);
    G_digits[G_tdigits++] = new Digit('5', d5);
    G_digits[G_tdigits++] = new Digit('6', d6);
    G_digits[G_tdigits++] = new Digit('7', d7);
    G_digits[G_tdigits++] = new Digit('8', d8);
    G_digits[G_tdigits++] = new Digit('9', d9);
    G_digits[G_tdigits++] = new Digit('-', dash);
    G_digits[G_tdigits++] = new Digit('/', slash);
    G_digits[G_tdigits++] = new Digit(':', dots);
    G_digits[G_tdigits++] = new Digit(' ', space);

    // UPDATE TIME
    UpdateTime();
    int len = strlen(G_dig);
    strcpy(G_lastdig, "-- -- -- --");

    // MAKE MAIN WINDOW
    int w = (FONTWIDTH + 2) * len + ( XMARGIN * 2 );
    int h = FONTHEIGHT + ( YMARGIN * 2 );
    G_mainwin = new MyWindow(w, h, "nixieclock");

#ifndef DARWIN
    // NON-BORDER WINDOWS DON'T WORK UNDER OSX YET
    G_mainwin->clear_border();
#endif

    G_mainwin->free_position();
    G_mainwin->show(argc,argv);

    // FLTK TIMER
    Fl::add_timeout(COMPHOLDSECS, Tick_CB);
    return(Fl::run());
}
